from django import template
from django.urls import reverse
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def webauthn_js(username_input, signup_ctrl, login_ctrl):
    code = '''<script> // django-webauthn initialization script
    window.initWebAuthnHandlers({
      credentialsCreate: "%(create)s",
      credentialsRegister: "%(register)s",
      credentialsGet: "%(get)s",
      verify: "%(verify)s",
      usernameInputSel: "%(user)s",
      signupSel: "%(signup)s",
      loginSel: "%(login)s",
    });
  </script>''' % {
        'create': reverse('webauthn:credentials-create'),
        'register': reverse('webauthn:credentials-register'),
        'get': reverse('webauthn:credentials-get'),
        'verify': reverse('webauthn:verify'),

        'user': username_input,
        'signup': signup_ctrl,
        'login': login_ctrl,
    }
    return mark_safe(code)



import json

from django.db import IntegrityError
from django.http import JsonResponse
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.views import View

from pywarp import RelyingPartyManager
from .conf import settings
from .forms import CreateOptionsForm, RegistrationForm, VerifyForm


class BaseView(View):
    @cached_property
    def rp(self):
        backend = import_string(settings.WEBAUTHN_BACKEND)()

        return RelyingPartyManager(
            settings.WEBAUTHN_NAME,
            rp_id=settings.WEBAUTHN_HOSTNAME,
            credential_storage_backend=backend,
            debug=True,
        )


class CredentialCreateOptionsView(BaseView):
    def post(self, request):
        form = CreateOptionsForm(json.loads(request.body))
        if not form.is_valid():
            return JsonResponse({
                'errors': form.errors.as_json(),
            })
        rp = self.rp
        data = form.cleaned_data
        try:
            reg_options = rp.get_registration_options(
                username=data['username'],
                full_name=data.get('full_name', None),
            )
        except IntegrityError:
            return JsonResponse({
                'errors': 'User "%s" already exists.' % data['username']
            })
        return JsonResponse(reg_options)


class RegisterCredentialView(BaseView):
    def post(self, request):
        form = RegistrationForm(json.loads(request.body))
        if not form.is_valid():
            return JsonResponse({
                'errors': form.errors.as_json()
            })
        rp = self.rp
        data = form.cleaned_data
        resp = rp.register(
            client_data_json=data.pop('client_data_json'),
            attestation_object=data.pop('attestation_object'),
            username=data.pop('username'),
            **data
        )
        return JsonResponse(resp)


class CredentialGetOptionsView(BaseView):
    def post(self, request):
        form = CreateOptionsForm(json.loads(request.body))
        if not form.is_valid():
            return JsonResponse({
                'errors': form.errors.as_json()
            })
        rp = self.rp
        data = rp.get_authentication_options(
            username=form.cleaned_data['username'],
        )
        return JsonResponse(data)


class VerifyAssertionView(BaseView):
    def post(self, request):
        form = VerifyForm(json.loads(request.body))
        if not form.is_valid():
            return JsonResponse({
                'errors': form.errors.as_json()
            })
        res = self.rp.verify(**form.cleaned_data)
        return JsonResponse(res)



from appconf import AppConf
from django.conf import settings


class WebAuthnConf(AppConf):
    name = 'webauthn'
    verbose_name = 'WebAuthn'

    # need to use settigns somehow to enable autoimport without flake8 complain
    NAME = 'WebAuthn App' + (' (debug)' if settings.DEBUG else '')

    HOSTNAME = 'localhost'
    BACKEND = 'webauthn.backends.ModelBackend'

    class Meta:
        prefix = 'webauthn'



import base64

from django import forms


class Base64Field(forms.CharField):

    def prepare_value(self, value):
        if value is not None:
            return base64.b64encode(value)

    def to_python(self, value):
        if value is not None:
            return base64.b64decode(value)


class CreateOptionsForm(forms.Form):
    username = forms.CharField()


class RegistrationForm(forms.Form):
    username = Base64Field()
    client_data_json = Base64Field()
    attestation_object = Base64Field()


class VerifyForm(forms.Form):
    username = Base64Field()
    authenticator_data = Base64Field()
    client_data_json = Base64Field()
    signature = Base64Field()
    user_handle = Base64Field()
    raw_id = forms.CharField(required=False)



from django.urls import path

from . import views

app_name = 'webauthn'

urlpatterns = [
    path('credentials/create/',
         views.CredentialCreateOptionsView.as_view(),
         name='credentials-create'),
    path('credentials/register/',
         views.RegisterCredentialView.as_view(),
         name='credentials-register'),
    path('credentials/get/',
         views.CredentialGetOptionsView.as_view(),
         name='credentials-get'),
    path('verify/',
         views.VerifyAssertionView.as_view(),
         name='verify'),
]



from django.contrib.auth import get_user_model

from pywarp import Credential
from pywarp.backends import AbstractStorageBackend
from webauthn.models import WebAuthnUser


class ModelBackend(AbstractStorageBackend):
    def __init__(self):
        self.user_model = get_user_model()

    def get_credential(self, username):
        user = self.user_model.objects.get_by_natural_key(username).webauthn
        return Credential(credential_id=user.credential_id,
                          credential_public_key=user.credential_public_key)

    def save_credential(self, username, credential, **user_extra):
        user = self.user_model.objects.get_by_natural_key(username).webauthn
        user.credential_id = credential.id
        user.credential_public_key = bytes(credential.public_key)
        user.save()

    def save_challenge(self, username, challenge, challenge_type):
        assert challenge_type in {"registration", "authentication"}

        if challenge_type == 'registration':
            WebAuthnUser.objects.create(
                user=self.user_model.objects.create(**{
                    self.user_model.USERNAME_FIELD: username
                }),
                registration_challenge=challenge,
            )
        else:
            user = self.user_model.objects.get_by_natural_key(username).webauthn
            user.authentication_challenge = challenge
            user.save()

    def get_challenge(self, username, challenge_type):
        assert challenge_type in {"registration", "authentication"}

        user = self.user_model.objects.get_by_natural_key(username).webauthn
        return getattr(user, challenge_type + "_challenge")



from django.db import models

from .conf import settings


class WebAuthnUser(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, models.CASCADE,
                                related_name='webauthn')
    credential_id = models.BinaryField(blank=True, null=True)
    credential_public_key = models.BinaryField(blank=True, null=True)
    registration_challenge = models.BinaryField(blank=True, null=True)
    authentication_challenge = models.BinaryField(blank=True, null=True)

    def __str__(self):
        return str(self.user_id)



import os

TEST_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'tests')

SECRET_KEY = '8q$p@8b$mbn$#7y)q2_dst^r-0l3d_ela!o413hka+#be$kc62'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'webauthn',
]

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
]
SESSION_SERIALIZER = 'django.contrib.sessions.serializers.PickleSerializer'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(TEST_DIR, 'db.sqlite3'),
    }
}

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import codecs
import os

from setuptools import find_packages, setup


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read().strip()


setup(
    name='django-webauthn',
    version=read('VERSION'),
    description="Django support for Web Authn.",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    keywords='django environment variables 12factor',
    author='Boris Shemigon',
    author_email='i@boris.co',
    url='https://github.com/shemigon/django-webauthn',
    license='MIT',
    packages=find_packages(),
    platforms=["any"],
    include_package_data=True,
    zip_safe=False,
    install_requires = [
        'django-appconf >= 1.0',
        'pywarp>=0.0.5',
    ],
    dependency_links=[
        'https://github.com/shemigon/pywarp/tarball/master#egg=pywarp-0.0.5',
    ],
    classifiers=[
        # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Framework :: Django'
    ]
)
